#define DECODER
#define BUFFER_NO 1
#define AUDIO_DEBUG 0
#define NCH 2
#define BPS 16
#define SAMPLERATE 44100

#include <iostream.h>
#include <math.h>
#include <string.h>
#include <fstream.h>
#include <stdlib.h>
#include "../filter_table.h"
#include <memory.h>
#include "../bitstream.h"
#include "../quant.h"

//Typendeklarationen
typedef struct
{
	unsigned int L[36];
	unsigned int R[36];
} QuantTyp;

typedef struct 
{
	unsigned int StreamVersion;
	unsigned int Bitrate;
	unsigned int Samplerate;
	unsigned int Frames;		// no of frames
	unsigned int Duration;		// in ms
	unsigned int VBR;
	unsigned int MS;
	unsigned int IS;
} MPEGplus_Stream_Info;

typedef struct
{
	unsigned int TAGpresent;
	char Title[31];
	char Artist[31];
	char Album[31];
	char Year[5];
	char Comment[31];
	char Genre[2];
} ID3TAG;


//interne Funktionen
void Requantisierung();
void Synthese_Filter(int);
void Calculate_New_V(float*,float*);
void Synthese_Filter_opt();
void Intensity_Stereo_Decode();
void Lese_Bitstrom();
void RESET_filter();

//externe Funktionen
void initiate_tables();
int file_init(char *fn);
int decode_frame(char *buffer, int *decoded_bits);
int get_fileinfo(char *fn, MPEGplus_Stream_Info *, ID3TAG *);
void perform_jump(int *eof, int desired_pos_ms);
void reset();
int write_tag( char *fn, ID3TAG *id3tag);



//globale Variablen
static float V_L[1024];
static float V_R[1024];
static float Y_L[36][32];
static float Y_R[36][32];
short Stream[2304];
unsigned int SCF_Index_L[3][32],SCF_Index_R[3][32];	//Skalenfaktor-Indices fuer Speicherung
QuantTyp Q[32];					//quantisierte Samples
unsigned int Res_L[32],Res_R[32];		//Aufloesungsstufen der Teilbaender
int DSCF_Flag_L[32],DSCF_Flag_R[32];	//Flag bei Benutzung von differential-SCF
unsigned int SCFI_L[32],SCFI_R[32];		//Uebertragungsreihenfolge der SCF
int DSCF_Reference_L[32],DSCF_Reference_R[32];

//globale Flags
int MS_used = 0;	//globaler Flag fuer M/S-Signalfuehrung
int MS_Flag[32];	//subbandweiser Flag fuer M/S-Signalfuehrung
int Max_used_Band=0;
int StreamVersion;
int BitstreamByte=0;
int Max_Band_desired=0;
int Blockgroesse;
int FwdJumpInfo;
int ActDecodePos;
unsigned int OverallFrames;
int N;


//--------------------
int file_length; // file length, in bytes
int decode_pos_ms; // current decoding position, in milliseconds
char sample_buffer[1152*NCH*(BPS/8)]; // sample buffer
//--------------------

void initiate_tables()
{
	/* any one-time initialization goes here (configuration reading, etc) */
	generiere_HuffmanTabelle();
	initialisiere_Filter();
}

void reset()
{
	// AB: resetting global variables
	dword=0;
	filled=0;
	pos=0;
	Zaehler=-1;
	BITS = 0;
	BitsRead = 0;
	DecodingEnd = 0;
	MaxByte = 0;
	MS_used = 0;
	IS_Flag = 0;
	//RESET_filter();
	memset(Y_L,0,36*32*4);memset(Y_R,0,36*32*4);
	memset(SCF_Index_L,0,3*32*4);memset(SCF_Index_R,0,3*32*4);
	memset(Res_L,0,32*4);memset(Res_R,0,32*4);
	memset(SCFI_L,0,32*4);memset(SCFI_R,0,32*4);
	memset(DSCF_Flag_L,0,32*4);memset(DSCF_Flag_R,0,32*4);
	memset(DSCF_Reference_L,0,32*4);memset(DSCF_Reference_R,0,32*4);
	memset(Q,0,32*72*4);
	memset(MS_Flag,0,32*4);

	// AB: resetting bitstream
	if (Speicher!=NULL)
	{
		free(Speicher);
		Speicher = NULL;
	}
}

int file_init(char *fn) 
{ 
	FILE *input_file;

	// open file
	input_file = fopen(fn, "rb");
	if (input_file==NULL) return 1;
	fseek(input_file, 0 ,SEEK_END);
	file_length = ftell(input_file);

	//AB: search for ID3-Tag and adapt filesize if necessary
	{
		FILE *InFileTmp;
		char tag[4];

		InFileTmp=fopen(fn, "rb");
		fseek( InFileTmp, -128, SEEK_END);
		fread((void*)&tag, 1, 3, InFileTmp);
	
		if (!strcmp(tag,"TAG")) file_length = file_length-128;
		
	}

	// AB: resetting all globals
	reset();

	// alloziere Speicher und lese Datei ein
	Speicher = (unsigned int*)calloc(file_length/sizeof(unsigned int),sizeof(unsigned int));
	MaxByte = file_length/sizeof(int);
	fseek( input_file, 0, SEEK_SET);
	fread((void*)Speicher, 1, file_length, input_file);

	// AB: decode fileheader
	RESET_filter();
	Bitrate  = (int)Bitstream_read(9);
	IS_Flag  = (int)Bitstream_read(1);
	MS_used  = (int)Bitstream_read(1);
	StreamVersion = (int)Bitstream_read(10);
	if (StreamVersion>=2) Max_Band_desired = (int)Bitstream_read(5);
	else Max_Band_desired = 0;
	if (StreamVersion>=3)
	{
		Blockgroesse = (int)Bitstream_read(6);
		if (StreamVersion<5) OverallFrames = (int)Bitstream_read(16);
		else                 OverallFrames = (int)Bitstream_read(32);
	}
	else
	{
		OverallFrames = 0;
		Blockgroesse = 1;
	}
	
	// AB: no valid file
	if (StreamVersion<1 && StreamVersion>5)
	{
		return 2;
	}

	initialisiere_Quantisierungstabellen(Max_Band_desired,1);

	decode_pos_ms=0;

	return 0; 
}

int get_fileinfo(char *fn, MPEGplus_Stream_Info *info, ID3TAG *id3tag)
{
	unsigned int HeaderData[2];
	FILE *tmpFile;
	unsigned int tmpStreamVersion,tmpBitrate,tmpMS,tmpIS,tmpFrames,tmpLength;

	//Lese File Header
	tmpFile = fopen(fn, "rb");
	fread((void*)&HeaderData, 1, 8, tmpFile);

	fseek(tmpFile, 0, SEEK_END);
	tmpLength = ftell(tmpFile);

	//Interpretiere File Header
	tmpBitrate       = (unsigned int) (HeaderData[0]>>23);
	tmpIS            = (unsigned int)((HeaderData[0]<<9 )>>31);
	tmpMS            = (unsigned int)((HeaderData[0]<<10)>>31);
	tmpStreamVersion = (unsigned int)((HeaderData[0]<<11)>>22);
	if (tmpStreamVersion>=3)
	{
		tmpFrames    = (unsigned int) HeaderData[1];
	}

	// Schreibe in struct
	info->StreamVersion = tmpStreamVersion;
	info->Samplerate = 44100;
	info->Frames = tmpFrames;
	info->Duration = (unsigned int)(tmpFrames*1152/44.1);
	info->MS = tmpMS;
	info->IS = tmpIS;
	if (tmpBitrate!=0)
	{
		info->VBR = 0;
		info->Bitrate = tmpBitrate;
	}
	else
	{
		info->VBR = 1;
		info->Bitrate = (unsigned int)(8.*tmpLength/info->Duration);
	}

	//Lese Tag-Info
	{
		FILE *InFileTmp;
		char tag[4]={0,0,0,0};

		memset(id3tag->Title  , 0, 31);
		memset(id3tag->Artist , 0, 31);
		memset(id3tag->Album  , 0, 31);
		memset(id3tag->Comment, 0, 31);
		memset(id3tag->Year   , 0, 5);
		memset(id3tag->Genre  , 0, 2);

		InFileTmp=fopen(fn, "rb");
		fseek( InFileTmp, -128, SEEK_END);
		fread((void*)&tag, 1, 3, InFileTmp);
	
		if (!strcmp(tag,"TAG")) id3tag->TAGpresent = 1;
		else					id3tag->TAGpresent = 0;

		if (id3tag->TAGpresent)
		{
			fread((void*)&id3tag->Title, 1, 30, InFileTmp);
			fread((void*)&id3tag->Artist, 1, 30, InFileTmp);
			fread((void*)&id3tag->Album, 1, 30, InFileTmp);
			fread((void*)&id3tag->Year, 1, 4, InFileTmp);
			fread((void*)&id3tag->Comment, 1, 30, InFileTmp); 
		}
		else
		{
			memcpy(id3tag->Title, fn, (strlen(fn)<31) ? strlen(fn) : 30 );
		}
		fclose(InFileTmp);
	}
	return 0;
}

int write_tag( char *fn, ID3TAG *id3tag)
{
	FILE *FileTmp;
	char tag[4] = {0,0,0,0};

	//Ueberpruefe, ob bereits Tag vorhanden
	FileTmp=fopen(fn, "rb");
	fseek( FileTmp, -128, SEEK_END);
	fread((void*)&tag, 1, 3, FileTmp);
	
	if (!strcmp(tag,"TAG")) id3tag->TAGpresent = 1;
	else					id3tag->TAGpresent = 0;
	

	//Schreibe Tag-Info
	FileTmp=fopen(fn, "rb+");
	if (id3tag->TAGpresent) fseek( FileTmp, -128+3, SEEK_END);
	else
	{
		char tag[4]="TAG";
		fseek( FileTmp, 0, SEEK_END); fwrite((void*)&tag, 1, 3, FileTmp);
	}

	fwrite((void*)&id3tag->Title, 1, 30, FileTmp);
	fwrite((void*)&id3tag->Artist, 1, 30, FileTmp);
	fwrite((void*)&id3tag->Album, 1, 30, FileTmp);
	fwrite((void*)&id3tag->Year, 1, 4, FileTmp);
	fwrite((void*)&id3tag->Comment, 1, 30, FileTmp);
	fwrite((void*)&id3tag->Genre, 1, 1, FileTmp);

	fclose(FileTmp);

	return 0;
}


void perform_jump(int *eof, int desired_pos_ms)
{
	int fwd = 0;
	int n;
	
	if (desired_pos_ms > (int)(OverallFrames/44.1*1152))
	{ 
		desired_pos_ms = (int)(OverallFrames/44.1*1152);
		*eof = 1;
	}
	fwd = (int)((desired_pos_ms - decode_pos_ms)*44.1/1152); //zu springende Frames
	if (StreamVersion<4)
	{
		if (fwd>0)
		{
			for (n=0; n<fwd; ++n) 
			{
				Lese_Bitstrom(); //Lese Bitstrom bis Sprungziel
				if (DecodingEnd) break;
				decode_pos_ms += (int)(1152/44.1);
			}
			if (!DecodingEnd) *eof = 0; else *eof = 1;
			RESET_filter();//resetting synthesis filter to avoid "clicks"
		}
	}
	else if (StreamVersion>=4)
	{
		int DesiredDecPos;
		int DesiredIndex;
		int DesiredBitPos;
		int JumpedFrames = 0;
		int n;
		float time_pos;
		if (fwd>0)
		{
			time_pos = (float)decode_pos_ms;
			//Gehe bis auf 32 Frames and das Sprungziel heran
			while (JumpedFrames < (fwd-32-Blockgroesse-1))
			{
				//calculate desired position
				DesiredDecPos = ActDecodePos + FwdJumpInfo;
				DesiredIndex  = DesiredDecPos/32;
				DesiredBitPos = DesiredDecPos%32;
				//set desired position
				Zaehler = DesiredIndex;
				pos = DesiredBitPos;
				//get next jump-info
				dword = Speicher[Zaehler];
				if (JumpedFrames+Blockgroesse < (fwd-32-Blockgroesse-1))
				{
					FwdJumpInfo = Bitstream_read(20);
					ActDecodePos = Zaehler*32 + pos;
				}
				JumpedFrames += Blockgroesse;
				time_pos += (float)(Blockgroesse*1152/44.1);
			}
			//Lese in den letzten (mindestens) 32 Frames vor dem Sprungziel den Bitstrom,
			//sodass die Skalenfaktoren korrekt vorliegen
			while(JumpedFrames < fwd)
			{
				for (n=0; n<Blockgroesse; ++n)
				{
					if ((JumpedFrames)%Blockgroesse==0)
					{
						FwdJumpInfo = Bitstream_read(20);
						ActDecodePos = Zaehler*32 + pos;
					}
					Lese_Bitstrom();
					if (DecodingEnd) break;
					++JumpedFrames;
				}
				time_pos += (float)(Blockgroesse*1152/44.1);
			}
			decode_pos_ms = (int)time_pos;
			if (!DecodingEnd) *eof = 0; else *eof = 1;
			RESET_filter();//resetting synthesis filter to avoid "clicks"
		}
		else
		{
			time_pos = (float)0; //Setze Zaehler auf Anfang
			fwd = (int)(desired_pos_ms*44.1/1152); //zu springende Frames
			
			Zaehler = -1; pos = 0;
			//dummy-read des stream-headers
			{
				int dummy;
				if (StreamVersion<5) dummy = Bitstream_read(48);
				else                 dummy = Bitstream_read(64);
			}

			//Nur Durchfuehren, falls Sprungziel _nicht_ Fileanfang ist
			//...ansonsten normales decoding ab file-header
			if (fwd>=32+Blockgroesse+1)
			{
				FwdJumpInfo = Bitstream_read(20);
				ActDecodePos = Zaehler*32+pos;
			
				//Gehe bis auf 32 Frames and das Sprungziel heran
				while (JumpedFrames < (fwd-32-Blockgroesse))
				{
					//calculate desired position
					DesiredDecPos = ActDecodePos + FwdJumpInfo;
					DesiredIndex  = DesiredDecPos/32;
					DesiredBitPos = DesiredDecPos%32;
					//set desired position
					Zaehler = DesiredIndex;
					pos = DesiredBitPos;
					//get next jump-info
					dword = Speicher[Zaehler];
					if (JumpedFrames+Blockgroesse<(fwd-32-Blockgroesse)) 
					{
						FwdJumpInfo = Bitstream_read(20);
						ActDecodePos = Zaehler*32 + pos;
					}
					JumpedFrames += Blockgroesse;
					time_pos += (float)(Blockgroesse*1152/44.1);
				}
				//Lese in den letzten (mindestens) 32 Frames vor dem Sprungziel den Bitstrom,
				//so dass die Skalenfaktoren korrekt vorliegen
				while(JumpedFrames < fwd)
				{
					for (n=0; n<Blockgroesse; ++n)
					{
						if ((JumpedFrames)%Blockgroesse==0)
						{
							FwdJumpInfo = Bitstream_read(20);
							ActDecodePos = Zaehler*32 + pos;
						}
						Lese_Bitstrom();
						++JumpedFrames;
					}
					time_pos += (float)(Blockgroesse*1152/44.1);
				}
			}
			decode_pos_ms = (int)(time_pos);
			*eof = 0;
			RESET_filter();//resetting synthesis filter to avoid "clicks"
		}
	}
}



void RESET_filter()
{
	memset(V_L,0,1024*sizeof(float));
	memset(V_R,0,1024*sizeof(float));
	memset(Stream,0,2304*sizeof(short));
	N = 0;
}


int decode_frame(char *buffer, int *decoded_bits)
{
	int bitposition = BitsRead;

	if (N%Blockgroesse==0 && StreamVersion>=4)
	{
		FwdJumpInfo = Bitstream_read(20);
		ActDecodePos = Zaehler*32 + pos;
	}
	Lese_Bitstrom();

	if (DecodingEnd) return 0;
			
	Requantisierung();
	if (IS_Flag) Intensity_Stereo_Decode();
	Synthese_Filter_opt();
		
	++N;
	*decoded_bits = BitsRead - bitposition;
	decode_pos_ms += (int)(1152/44.1);

	memcpy(buffer,Stream,4608);
	return 4608;
}

void Calculate_New_V(float *Sample,float *V)
{

  //Calculating new V-buffer values for left channel
  //Berechnen neuer V-Werte (s. ISO-11172-3 S. 39)
  //basierend auf dem Algorithmus von Byeong Gi Lee
  
  int i,k,n;
  float tmp;
  float *C=&Cos64[0];
  static float Ai[16],Bi[16];

  Ai[i=0] = Sample[k=0] + Sample[n=31];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  Ai[++i] = Sample[++k] + Sample[--n];
  

  Bi[i=0] =  Ai[k=0] + Ai[n=15];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[k=0] - Ai[n=15])* C[2];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[6];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[10];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[14];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[18];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[22];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[26];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[30];
 
  Ai[i=0] =  Bi[k=0] + Bi[n=7];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] = (Bi[k=0] - Bi[n=7]) * C[4];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[12];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[20];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[28];
  Ai[++i] =  Bi[k=8] + Bi[n=15];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] = (Bi[k=8] - Bi[n=15])* C[4];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[12];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[20];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[28];

  tmp = C[8];
  Bi[i=0] =  Ai[k=0] + Ai[n=3];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=4] + Ai[n=7];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=8] + Ai[n=11];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=12] + Ai[n=15];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];

  tmp = C[16];
  Ai[i=0] =  Bi[k=0] + Bi[n=1];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;

  V[12]=   Ai[7];
  tmp = Ai[5] + Ai[7];
  V[4] =   tmp;
  V[36]= -(tmp + Ai[6]);
  V[44]= - Ai[4] - Ai[6] - Ai[7];
  V[14]=   Ai[15];
  tmp = Ai[11] + Ai[15];
  V[10]=   tmp;
  V[6] =   tmp + Ai[13];
  tmp = Ai[9] + Ai[13] + Ai[15];
  V[2] =   tmp;
  V[34]= -(tmp + Ai[14]);
  V[38]= -(tmp - Ai[9] + Ai[10] + Ai[11] + Ai[14]);
  tmp = Ai[12] + Ai[14] + Ai[15];
  V[46]= -(tmp + Ai[8]);
  V[42]= -(tmp + Ai[10] + Ai[11]);
  V[48]= - Ai[0];
  V[0] =   Ai[1];
  V[8] =   Ai[3];
  V[40]= - Ai[2] - Ai[3];
 
  Ai[i=0] = (Sample[k=0] - Sample[n=31])* C[1];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[3];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[5];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[7];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[9];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[11];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[13];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[15];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[17];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[19];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[21];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[23];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[25];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[27];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[29];
  Ai[++i] = (Sample[++k] - Sample[--n]) * C[31];
 
  Bi[i=0] =  Ai[k=0] + Ai[n=15];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[k=0] - Ai[n=15])* C[2];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[6];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[10];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[14];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[18];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[22];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[26];
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[30];

  Ai[i=0] =  Bi[k=0] + Bi[n=7];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] = (Bi[k=0] - Bi[n=7]) * C[4];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[12];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[20];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[28];
  Ai[++i] =  Bi[k=8] + Bi[n=15];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] =  Bi[++k] + Bi[--n];
  Ai[++i] = (Bi[k=8] - Bi[n=15])* C[4];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[12];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[20];
  Ai[++i] = (Bi[++k] - Bi[--n]) * C[28];

  tmp = C[8];
  Bi[i=0] =  Ai[k=0] + Ai[n=3];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=4] + Ai[n=7];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=8] + Ai[n=11];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];
  Bi[++i] =  Ai[k=12] + Ai[n=15];
  Bi[++i] =  Ai[++k] + Ai[--n];
  Bi[++i] = (Ai[--k] - Ai[++n]) * tmp;
  Bi[++i] = (Ai[++k] - Ai[--n]) * C[24];

  tmp = C[16];
  Ai[i=0] =  Bi[k=0] + Bi[n=1];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;
  Ai[++i] =  Bi[++k] + Bi[++n];
  Ai[++i] = (Bi[k++] - Bi[n++]) * tmp;

  V[15]= Ai[15];
  tmp = Ai[7] + Ai[15];
  V[13]= tmp;
  tmp += Ai[11];
  V[5] = tmp + Ai[5] + Ai[13];
  V[11]= tmp;
  tmp = Ai[3] + Ai[11] + Ai[15];
  V[7] = tmp + Ai[13];
  V[9] = tmp;
  tmp = Ai[1] + Ai[9] + Ai[13] + Ai[15];
  V[1] = tmp;
  V[33]= -(tmp + Ai[14]);
  tmp = Ai[5] + Ai[7] + Ai[9] + Ai[13] + Ai[15];
  V[3] = tmp;
  V[35]= -(tmp + Ai[6] +Ai[14]);
  tmp = Ai[10] + Ai[11] + Ai[13] + Ai[14] + Ai[15];
  V[37]= -(tmp + Ai[5] + Ai[6] + Ai[7]);
  V[39]= -(tmp + Ai[2] + Ai[3]);
  tmp = tmp + Ai[12] - Ai[13];;
  V[41]= -(tmp + Ai[2] + Ai[3]);
  V[43]= -(tmp + Ai[4] + Ai[6] + Ai[7]);

  tmp = Ai[8] + Ai[12] + Ai[14] + Ai[15];
  V[47]= -(tmp + Ai[0]);
  V[45]= -(tmp + Ai[4] + Ai[6] + Ai[7]);

    V[32] = -V[0];
	V[31] = -V[1];
	V[30] = -V[2];
	V[29] = -V[3];
	V[28] = -V[4];
	V[27] = -V[5];
	V[26] = -V[6];
	V[25] = -V[7];
	V[24] = -V[8];
	V[23] = -V[9];
	V[22] = -V[10];
	V[21] = -V[11];
	V[20] = -V[12];
	V[19] = -V[13];
	V[18] = -V[14];
	V[17] = -V[15];

	V[63] = V[33];
	V[62] = V[34];
	V[61] = V[35];
	V[60] = V[36];
	V[59] = V[37];
	V[58] = V[38];
	V[57] = V[39];
	V[56] = V[40];
	V[55] = V[41];
	V[54] = V[42];
	V[53] = V[43];
	V[52] = V[44];
	V[51] = V[45];
	V[50] = V[46];
	V[49] = V[47];
}

void Synthese_Filter_opt()
{
	int n,k,Index1;
	int Sum;
	
	for (n=0; n<36; ++n)
	{
		//shifting um 64 Indices aufwaerts
		memmove(V_L+64,V_L,3840);

		Calculate_New_V(&Y_L[n][0],&V_L[0]);

		//vectoring & windowing & calculating PCM-Output
		for (k=0; k<32; ++k)
		{
			Index1 = k*16;
			Sum=(int)(V_L[k+992]*Di_opt[Index1+15]
					+ V_L[k+896]*Di_opt[Index1+14]
					+ V_L[k+864]*Di_opt[Index1+13]
					+ V_L[k+768]*Di_opt[Index1+12]
					+ V_L[k+736]*Di_opt[Index1+11]
					+ V_L[k+640]*Di_opt[Index1+10]
					+ V_L[k+608]*Di_opt[Index1+ 9]
					+ V_L[k+512]*Di_opt[Index1+ 8]
					+ V_L[k+480]*Di_opt[Index1+ 7]
					+ V_L[k+384]*Di_opt[Index1+ 6]
					+ V_L[k+352]*Di_opt[Index1+ 5]
					+ V_L[k+256]*Di_opt[Index1+ 4]
					+ V_L[k+224]*Di_opt[Index1+ 3]
					+ V_L[k+128]*Di_opt[Index1+ 2]
					+ V_L[k+ 96]*Di_opt[Index1+ 1]
					+ V_L[k    ]*Di_opt[Index1   ]);

			//copy to PCM
			Stream[n*64+k*2] = Sum;
			//prevent from overflow
			if      (Sum> 32767) Stream[n*64+k*2] =  32767;
			else if (Sum<-32768) Stream[n*64+k*2] = -32768;
		}
	}
	for (n=0; n<36; ++n)
	{
		//shifting um 64 Indices aufwaerts
		memmove(V_R+64,V_R,3840);

		Calculate_New_V(&Y_R[n][0],&V_R[0]);

		//vectoring & windowing & calculating PCM-Output
		for (k=0; k<32; ++k)
		{
			Index1 = k*16;
			Sum=(int)(V_R[k+992]*Di_opt[Index1+15]
					+ V_R[k+896]*Di_opt[Index1+14]
					+ V_R[k+864]*Di_opt[Index1+13]
					+ V_R[k+768]*Di_opt[Index1+12]
					+ V_R[k+736]*Di_opt[Index1+11]
					+ V_R[k+640]*Di_opt[Index1+10]
					+ V_R[k+608]*Di_opt[Index1+ 9]
					+ V_R[k+512]*Di_opt[Index1+ 8]
					+ V_R[k+480]*Di_opt[Index1+ 7]
					+ V_R[k+384]*Di_opt[Index1+ 6]
					+ V_R[k+352]*Di_opt[Index1+ 5]
					+ V_R[k+256]*Di_opt[Index1+ 4]
					+ V_R[k+224]*Di_opt[Index1+ 3]
					+ V_R[k+128]*Di_opt[Index1+ 2]
					+ V_R[k+ 96]*Di_opt[Index1+ 1]
					+ V_R[k    ]*Di_opt[Index1   ]);

			//copy to PCM
			Stream[n*64+k*2+1] = Sum;
			//prevent from overflow
			if      (Sum> 32767) Stream[n*64+k*2+1] =  32767;
			else if (Sum<-32768) Stream[n*64+k*2+1] = -32768;
		}
	}
}

void Requantisierung()
{
	int Band,Last_Band,Subframe,n;

	//Requantisierung und Skalierung der Subband- und Subframesamples
	if (!IS_Flag)	Last_Band = Max_Band;
	else			Last_Band = Min_Band-1;

	for (Band=0; Band<=Last_Band; ++Band)
	{
		float templ,tempr;

		if (MS_Flag[Band]==1)
		{
			if      (Res_L[Band]!=0 && Res_R[Band]==0) for (n=0; n<36; ++n)
			{
				Y_R[n][Band] = Y_L[n][Band] = Requantisieren(Q[Band].L[n],Res_L[Band])*SCF[SCF_Index_L[n/12][Band]];
			}
			else if (Res_L[Band]!=0 && Res_R[Band]!=0) for (n=0; n<36; ++n)
			{
				Subframe = n/12;
				templ = Requantisieren(Q[Band].L[n],Res_L[Band])*SCF[SCF_Index_L[Subframe][Band]];
				tempr = Requantisieren(Q[Band].R[n],Res_R[Band])*SCF[SCF_Index_R[Subframe][Band]];
				Y_L[n][Band] = (templ+tempr);
				Y_R[n][Band] = (templ-tempr);
			}
			else if (Res_L[Band]==0 && Res_R[Band]!=0) for (n=0; n<36; ++n)
			{
				Y_L[n][Band] = Requantisieren(Q[Band].R[n],Res_R[Band])*SCF[SCF_Index_R[n/12][Band]];
				Y_R[n][Band] = -Y_L[n][Band];
			}
			else for (n=0; n<36; ++n) Y_L[n][Band] = Y_R[n][Band] = (float)(0);
		}
		else /*if (MS_Flag[Band]==0)*/
		{
			if      (Res_L[Band]!=0 && Res_R[Band]!=0) for (n=0; n<36; ++n)
			{
				Subframe = n/12;
				Y_L[n][Band] = Requantisieren(Q[Band].L[n],Res_L[Band])*SCF[SCF_Index_L[Subframe][Band]];
				Y_R[n][Band] = Requantisieren(Q[Band].R[n],Res_R[Band])*SCF[SCF_Index_R[Subframe][Band]];
			}
			else if (Res_L[Band]==0 && Res_R[Band]!=0) for (n=0; n<36; ++n)
			{
				Y_L[n][Band] = (float)(0);
				Y_R[n][Band] = Requantisieren(Q[Band].R[n],Res_R[Band])*SCF[SCF_Index_R[n/12][Band]];
			}
			else if (Res_L[Band]!=0 && Res_R[Band]==0) for (n=0; n<36; ++n)
			{
				Y_L[n][Band] = Requantisieren(Q[Band].L[n],Res_L[Band])*SCF[SCF_Index_L[n/12][Band]];
				Y_R[n][Band] = (float)(0);
			}
			else  for (n=0; n<36; ++n) Y_L[n][Band] = Y_R[n][Band] = (float)(0);
		}
	}
}

void Intensity_Stereo_Decode()
{
	int Band,n,k;
	float norm;

	//Requantisierung und Wiederherstellung des Stereo-Images
	norm = (float)(sqrt(sqrt(2)));
	for (Band=Min_Band; Band<=Max_Band; ++Band)
	{
		for (n=0; n<36; ++n)
		{
			if (Res_L[Band]!=0)
			{
				k=n/12;
				
				Y_L[n][Band] = Requantisieren(Q[Band].L[n],Res_L[Band])/norm;

				Y_R[n][Band] = Y_L[n][Band]*SCF[SCF_Index_R[k][Band]];
				Y_L[n][Band] = Y_L[n][Band]*SCF[SCF_Index_L[k][Band]];
			}
			else
			{
				Y_L[n][Band] = Y_R[n][Band] = (float)(0);
			}
		}
	}
}

void Lese_Bitstrom()
{
	int n,k;
	HuffmanTyp *Table;
	Max_used_Band=0;

	//reading Frameheader
	for (n=0; n<=Max_Band; ++n)
	{
		if      (n<11)           Table = &Region_A[0];
		else if (n>=11 && n<=22) Table = &Region_B[0];
		else /*if (n>=23)*/      Table = &Region_C[0];

		if (Bitrate<=128) Res_L[n] = Q_res[n][Huffman_Decode(Table)];
		else			  Res_L[n] = Q_res[n][Bitstream_read(Q_bit[n])];

		//nur Lesen, falls kein IS
		if (!(IS_Flag==1 && n>=Min_Band))
		{
			if (MS_used)	  MS_Flag[n] = Bitstream_read(1);
			if (Bitrate<=128) Res_R[n] = Q_res[n][Huffman_Decode(Table)];
			else			  Res_R[n] = Q_res[n][Bitstream_read(Q_bit[n])];
		}
		else
		{
			Res_R[n]=0;
		}
		//fuehre nachfolgende Operationen nur bis zum maximal enthaltenen Subband aus
		if (Res_L[n]!=0 || Res_R[n]!=0) Max_used_Band = n;
	}
			
	//reading SCFI-Bundle
	for (n=0; n<=Max_used_Band; ++n)
	{
		if (Res_L[n]>0)
		{
			BundleTyp Word;
			Word = SCFI_Bundle_read();
			SCFI_L[n] = Word.SCFI;
			DSCF_Flag_L[n] = Word.DSCF;
		}
		if ((Res_R[n]>0) || (Res_L[n]>0 && IS_Flag==1 && n>=Min_Band))
		{
			BundleTyp Word;
			Word = SCFI_Bundle_read();
			SCFI_R[n] = Word.SCFI;
			DSCF_Flag_R[n] = Word.DSCF;
		}
	}
	
	//reading SCF
	Table = &DSCF_Entropie[0];
	for (n=0; n<=Max_used_Band; ++n)
	{
		if (Res_L[n]>0)
		{   
			//mit DSCF
			if (DSCF_Flag_L[n]==1)
			{
				if      (SCFI_L[n]==3)
				{
					SCF_Index_L[0][n] = DSCF_Reference_L[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[1][n] = SCF_Index_L[0][n];
					SCF_Index_L[2][n] = SCF_Index_L[1][n];
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else if (SCFI_L[n]==1)
				{
					SCF_Index_L[0][n] = DSCF_Reference_L[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[1][n] = SCF_Index_L[0][n]   + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[2][n] = SCF_Index_L[1][n];
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else if (SCFI_L[n]==2)
				{
					SCF_Index_L[0][n] = DSCF_Reference_L[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[1][n] = SCF_Index_L[0][n];
					SCF_Index_L[2][n] = SCF_Index_L[1][n]   + Huffman_Decode_fast(Table) -6;
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else /*if      (SCFI_L[n]==0)*/
				{
					SCF_Index_L[0][n] = DSCF_Reference_L[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[1][n] = SCF_Index_L[0][n]   + Huffman_Decode_fast(Table) -6;
					SCF_Index_L[2][n] = SCF_Index_L[1][n]   + Huffman_Decode_fast(Table) -6;
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
			}
			//ohne DSCF
			else /*if (DSCF_Flag_L[n]==0)*/
			{
				if      (SCFI_L[n]==3)
				{
					SCF_Index_L[0][n] = Bitstream_read(6);
					SCF_Index_L[1][n] = SCF_Index_L[0][n];
					SCF_Index_L[2][n] = SCF_Index_L[1][n];
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else if (SCFI_L[n]==1)
				{
					SCF_Index_L[0][n] = Bitstream_read(6);
					SCF_Index_L[1][n] = Bitstream_read(6);
					SCF_Index_L[2][n] = SCF_Index_L[1][n];
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else if (SCFI_L[n]==2)
				{
					SCF_Index_L[0][n] = Bitstream_read(6);
					SCF_Index_L[1][n] = SCF_Index_L[0][n];
					SCF_Index_L[2][n] = Bitstream_read(6);
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
				else /*if      (SCFI_L[n]==0)*/
				{
					SCF_Index_L[0][n] = Bitstream_read(6);
					SCF_Index_L[1][n] = Bitstream_read(6);
					SCF_Index_L[2][n] = Bitstream_read(6);
					DSCF_Reference_L[n] = SCF_Index_L[2][n];
				}
			}
		}
		if ((Res_R[n]>0) || (Res_L[n]>0 && IS_Flag==1 && n>=Min_Band))
		{
			//mit DSCF
			if (DSCF_Flag_R[n]==1)
			{
				if      (SCFI_R[n]==3)
				{
					SCF_Index_R[0][n] = DSCF_Reference_R[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[1][n] = SCF_Index_R[0][n];
					SCF_Index_R[2][n] = SCF_Index_R[1][n];
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else if (SCFI_R[n]==1)
				{
					SCF_Index_R[0][n] = DSCF_Reference_R[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[1][n] = SCF_Index_R[0][n]   + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[2][n] = SCF_Index_R[1][n];
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else if (SCFI_R[n]==2)
				{
					SCF_Index_R[0][n] = DSCF_Reference_R[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[1][n] = SCF_Index_R[0][n];
					SCF_Index_R[2][n] = SCF_Index_R[1][n]   + Huffman_Decode_fast(Table) -6;
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else /*if (SCFI_R[n]==0)*/
				{
					SCF_Index_R[0][n] = DSCF_Reference_R[n] + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[1][n] = SCF_Index_R[0][n]   + Huffman_Decode_fast(Table) -6;
					SCF_Index_R[2][n] = SCF_Index_R[1][n]   + Huffman_Decode_fast(Table) -6;
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
			}
			//ohne DSCF
			else /*if (DSCF_Flag_R[n]==0)*/
			{
				if      (SCFI_R[n]==3)
				{
					SCF_Index_R[0][n] = Bitstream_read(6);
					SCF_Index_R[1][n] = SCF_Index_R[0][n];
					SCF_Index_R[2][n] = SCF_Index_R[1][n];
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else if (SCFI_R[n]==1)
				{
					SCF_Index_R[0][n] = Bitstream_read(6);
					SCF_Index_R[1][n] = Bitstream_read(6);
					SCF_Index_R[2][n] = SCF_Index_R[1][n];
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else if (SCFI_R[n]==2)
				{
					SCF_Index_R[0][n] = Bitstream_read(6);
					SCF_Index_R[1][n] = SCF_Index_R[0][n];
					SCF_Index_R[2][n] = Bitstream_read(6);
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
				else /*if      (SCFI_R[n]==0)*/
				{
					SCF_Index_R[0][n] = Bitstream_read(6);
					SCF_Index_R[1][n] = Bitstream_read(6);
					SCF_Index_R[2][n] = Bitstream_read(6);
					DSCF_Reference_R[n] = SCF_Index_R[2][n];
				}
			}
		}
	}

	//reading Samples
	for (n=0; n<=Max_used_Band; ++n)
	{
		HuffmanTyp *x1,*x2;
		
		if      (Res_L[n] == 1) {x1 = &Entropie_1[0];}
		else if (Res_L[n] == 2) {x1 = &Entropie_2[0];}
		else if (Res_L[n] == 3) {x1 = &Entropie_3[0];}
		else if (Res_L[n] == 4) {x1 = &Entropie_4[0];}
		else if (Res_L[n] == 5) {x1 = &Entropie_5[0];}
		else if (Res_L[n] == 6) {x1 = &Entropie_6[0];}
		else if (Res_L[n] == 7) {x1 = &Entropie_7[0];}
		else					{x1 = NULL;}

		if      (Res_R[n] == 1) {x2 = &Entropie_1[0];}
		else if (Res_R[n] == 2) {x2 = &Entropie_2[0];}
		else if (Res_R[n] == 3) {x2 = &Entropie_3[0];}
		else if (Res_R[n] == 4) {x2 = &Entropie_4[0];}
		else if (Res_R[n] == 5) {x2 = &Entropie_5[0];}
		else if (Res_R[n] == 6) {x2 = &Entropie_6[0];}
		else if (Res_R[n] == 7) {x2 = &Entropie_7[0];}
		else					{x2 = NULL;}

		for (k=0; k<36; ++k)
		{
			if (x1!=NULL) Q[n].L[k] = Huffman_Decode_fast(x1);
			if (x2!=NULL) Q[n].R[k] = Huffman_Decode_fast(x2);
		}
		
		if (!(Res_L[n]<=7 && Res_R[n]<=7))
		for (k=0; k<36; ++k)
		{
			if (Res_L[n]>7) {Q[n].L[k] = Bitstream_read(Res_bit[Res_L[n]]);}
			if (Res_R[n]>7) {Q[n].R[k] = Bitstream_read(Res_bit[Res_R[n]]);}
		}
	}
}





